@@ -10,14 +10,16 @@ use ChemOnomatopist;
 use ChemOnomatopist::Chain::Amide;
 use ChemOnomatopist::Chain::Amine;
 use ChemOnomatopist::Chain::Ether;
+use ChemOnomatopist::Charge;
 use ChemOnomatopist::Elements qw( %elements );
 use ChemOnomatopist::Group::Carboxyl;
 use ChemOnomatopist::Group::Ether;
 use ChemOnomatopist::Isotope;
 use ChemOnomatopist::Name::Part::AlkaneANSuffix;
+use ChemOnomatopist::Name::Part::Isotope;
 use ChemOnomatopist::Util::SMILES qw( path_SMILES );
 use Graph::Traversal::DFS;
-use List::Util qw( all any none sum0 uniq );
+use List::Util qw( all any first none sum0 uniq );
 use Scalar::Util qw( blessed );
 use Set::Object qw( set );
 
@@ -62,6 +64,17 @@ sub parent(;$)
     return $old_parent;
 }
 
+sub parent_locant()
+{
+    my( $self ) = @_;
+    my $parent = $self->parent;
+    return $parent unless $parent;
+
+    my $graph = $self->graph;
+    my @vertices = $self->vertices;
+    return first { $graph->has_edge( $vertices[$_], $parent ) } 0..$#vertices;
+}
+
 sub substituents()
 {
     my( $self ) = @_;
@@ -185,8 +198,8 @@ sub nonstandard_valence_positions()
     my @nonstandard_valence_positions;
     for (0..$#vertices) {
         next if blessed $vertices[$_];
-        next if ChemOnomatopist::is_element( $vertices[$_], 'C' );
-        next if ChemOnomatopist::is_element( $vertices[$_], 'N' );
+        next if ChemOnomatopist::element( $vertices[$_] ) eq 'C';
+        next if ChemOnomatopist::element( $vertices[$_] ) eq 'N';
         next unless exists $vertices[$_]->{valence};
         push @nonstandard_valence_positions, $_;
     }
@@ -285,6 +298,15 @@ sub double_bond_positions()
     return grep { $bonds[$_] eq '=' } 0..$#bonds;
 }
 
+# FIXME: Needs proper name
+sub needs_ane_suffix()
+{
+    my( $self ) = @_;
+    return $self->length == 1 && blessed $self->{vertices}[0] &&
+           ($self->{vertices}[0]->isa( ChemOnomatopist::Group::Sulfinyl:: ) ||
+            $self->{vertices}[0]->isa( ChemOnomatopist::Group::Sulfonyl:: ));
+}
+
 sub needs_multiple_bond_locants()
 {
     my( $self ) = @_;
@@ -330,7 +352,7 @@ sub needs_heteroatom_names()
     }
 
     # Chalcogen analogues of ethers
-    if( @vertices == 1 && grep { ChemOnomatopist::element( @vertices ) eq $_ } qw( S Se Te ) ) {
+    if( @vertices == 1 && $self->parent && grep { ChemOnomatopist::element( @vertices ) eq $_ } qw( S Se Te ) ) {
         return '';
     }
 
@@ -352,7 +374,9 @@ sub needs_suffix_locant()
     my @most_senior_groups = $self->most_senior_groups;
     return '' unless @most_senior_groups;
     return 1 if $self->number_of_heteroatoms; # P-15.4.3.2.3: Characteristic groups cited as suffixes are given locants
-    return 1 if !$most_senior_groups[0]->is_carbon || @most_senior_groups > 2;
+    return 1 if @most_senior_groups > 2;
+    return 1 if !ChemOnomatopist::element( $most_senior_groups[0] );
+    return 1 if  ChemOnomatopist::element( $most_senior_groups[0] ) ne 'C';
     return '';
 }
 
@@ -361,6 +385,8 @@ sub needs_substituent_locants()
     my( $self ) = @_;
     return '' if $self->length == 1;
 
+    return 1 if $self->number_of_isotopes;
+
     # FIXME: Make sure the substituents are of the same kind
     return '' if scalar( $self->substituents ) == $self->max_valence;
 
@@ -374,8 +400,28 @@ sub needs_substituent_locants()
     return 1;
 }
 
+sub needs_charge_locants()  { &needs_substituent_locants }
 sub needs_isotope_locants() { &needs_substituent_locants }
 
+sub charges()
+{
+    my( $self ) = @_;
+    return @{$self->{charges}} if $self->{charges};
+
+    my @vertices = $self->vertices;
+    my @charges;
+    for my $i (0..$#vertices) {
+        next if blessed $vertices[$i];
+        next unless $vertices[$i]->{charge};
+        push @charges, ChemOnomatopist::Charge->new( $vertices[$i]->{charge},
+                                                     $i,
+                                                     $self->locants( $i ) );
+    }
+
+    $self->{charges} = \@charges;
+    return @charges;
+}
+
 sub heteroatoms()
 {
     my( $self ) = @_;
@@ -506,6 +552,12 @@ sub number_of_carbons()
     return $C;
 }
 
+sub number_of_charges()
+{
+    my( $self ) = @_;
+    return scalar $self->charges;
+}
+
 sub number_of_branches()
 {
     my( $self ) = @_;
@@ -554,6 +606,39 @@ sub number_of_nonstandard_valence_positions()
     return scalar $self->nonstandard_valence_positions;
 }
 
+sub charge_part()
+{
+    my( $self ) = @_;
+    return ChemOnomatopist::Name->new unless $self->charges;
+
+    my @negative = grep { $_->charge < 0 } $self->charges;
+    my @positive = grep { $_->charge > 0 } $self->charges;
+
+    my $name = ChemOnomatopist::Name->new;
+
+    if( @positive ) {
+        @positive = map { ( $_->locant ) x abs $_->charge } @positive;
+        $name->append_locants( @positive ) if $self->needs_charge_locants;
+        if( @positive > 1 ) {
+            $name->append_multiplier( ChemOnomatopist::IUPAC_complex_numerical_multiplier( scalar @positive ) );
+            $name .= '(';
+        }
+        $name .= @negative ? 'ium' : 'ylium';
+        $name .= ')' if @positive > 1;
+    }
+
+    if( @negative ) {
+        @negative = map { ( $_->locant ) x abs $_->charge } @negative;
+        $name->append_locants( @negative ) if $self->needs_charge_locants;
+        if( @negative > 1 ) {
+            $name->append_multiplier( ChemOnomatopist::IUPAC_numerical_multiplier( scalar @negative ) );
+        }
+        $name .= 'ide';
+    }
+
+    return $name;
+}
+
 sub indicated_hydrogens_part() { ChemOnomatopist::Name->new }
 
 sub isotope_part()
@@ -561,14 +646,14 @@ sub isotope_part()
     my( $self ) = @_;
 
     my @isotopes = sort { $a->element cmp $b->element ||
-                          $a->atomic_number <=> $b->atomic_number }
+                          $a->mass_number <=> $b->mass_number }
                         $self->isotopes;
     return '' unless @isotopes;
 
     my @order;
     my %freq;
     for my $isotope (@isotopes) {
-        my $key = $isotope->atomic_number . $isotope->element;
+        my $key = $isotope->mass_number . $isotope->element;
         if( !$freq{$key} ) {
             $freq{$key} = [];
             push @order, $key;
@@ -589,21 +674,28 @@ sub isotope_part()
         }
     }
 
-    return "($isotopes)";
+    return ChemOnomatopist::Name::Part::Isotope->new( "($isotopes)" );
 }
 
 sub prefix()
 {
     my( $self ) = @_;
 
-    # Chalcogen analogues of ethers
     if( $self->length == 1 ) {
-        return ChemOnomatopist::Name->new( 'sulfan' ) if ChemOnomatopist::is_element( $self->vertices, 'S' );
-        return ChemOnomatopist::Name->new( 'selan'  ) if ChemOnomatopist::is_element( $self->vertices, 'Se' );
-        return ChemOnomatopist::Name->new( 'tellan' ) if ChemOnomatopist::is_element( $self->vertices, 'Te' );
+        my $vertex = $self->{vertices}[0];
+        return $vertex->prefix if blessed $vertex;
+
+        # Chalcogen analogues of ethers
+        my $element = ChemOnomatopist::element( $vertex );
+        return ChemOnomatopist::Name->new( 'sulfan' ) if $element eq 'S';
+        return ChemOnomatopist::Name->new( 'selan'  ) if $element eq 'Se';
+        return ChemOnomatopist::Name->new( 'tellan' ) if $element eq 'Te';
     }
 
-    return $self->suffix; # FIXME: Add proper suffix
+    my $name = $self->suffix;
+    $name->pop_e;
+    pop @$name if $name->ends_with_alkane_an_suffix;
+    return $name . 'yl';
 }
 
 sub suffix()
@@ -613,18 +705,20 @@ sub suffix()
     my @chain = $self->vertices;
 
     my $name = ChemOnomatopist::Name->new;
-    if( $self->length == 1 && !blessed $chain[0] && !ChemOnomatopist::is_element( @chain, 'C' ) ) {
-        $name .= 'ne'; # Leaving element prefix appending to the caller
-        return $name;
+    if( $self->length == 1 && !blessed $chain[0] && ChemOnomatopist::element( @chain ) ne 'C' ) {
+        return $name . 'ne'; # Leaving element prefix appending to the caller
     }
 
+    # CHECKME: Not sure if calling prefix() is correct
+    return $chain[0]->prefix if $self->length == 1 && blessed $chain[0];
+
     my @bonds = $self->bonds;
     my @double = grep { $bonds[$_] eq '=' } 0..$#bonds;
     my @triple = grep { $bonds[$_] eq '#' } 0..$#bonds;
 
     # BBv2 P-63.2.2.2
-    if( $self->parent && (all { !blessed $_ } @chain) && ChemOnomatopist::is_element( $chain[0], 'O' ) &&
-        !@double && !@triple && all { ChemOnomatopist::is_element( $_, 'C' ) } @chain[1..$#chain] ) {
+    if( $self->parent && @chain && (all { !blessed $_ } @chain) && ChemOnomatopist::element( @chain ) eq 'O' &&
+        !@double && !@triple && all { ChemOnomatopist::element( $_ ) eq 'C' } @chain[1..$#chain] ) {
         $name->append_stem( ChemOnomatopist::alkane_chain_name( $self->length - 1 ) );
         $name .= 'oxy';
         return $name;
