@@ -11,10 +11,12 @@ use ChemOnomatopist::Chain::Circular;
 use ChemOnomatopist::Chain::Bicycle::Purine;
 use ChemOnomatopist::Chain::Monocycle;
 use ChemOnomatopist::Chain::Monocycle::Fused;
+use ChemOnomatopist::Comparable::Array::Numeric;
 use ChemOnomatopist::Elements qw( %elements );
 use ChemOnomatopist::Name;
 use ChemOnomatopist::Name::Part::Fusion;
 use ChemOnomatopist::Name::Part::Stem;
+use ChemOnomatopist::Util qw( all_max all_min );
 use ChemOnomatopist::Util::SMILES qw( cycle_SMILES );
 use Chemistry::OpenSMILES qw( is_double_bond );
 use Graph::Traversal::DFS;
@@ -26,14 +28,13 @@ use parent ChemOnomatopist::Chain::Circular::;
 
 # From BBv2 P-25.2.1
 our @names = (
-    [ 'n:c:n:c:c:c:', 'n:c:c:n:c:c:', 'pteridine' ],
-    [ 'c:n:c:n:c:c:', 'n:c:c:n:c:c:', 'pteridine' ],
+    [ qw( NCNCCC NCCNCC pteridine ) ],
+    [ qw( CNCNCC NCCNCC pteridine ) ],
 
-    [ 'n:n:c:c:c:c:', 'c:c:c:c:c:c:', 'cinnoline' ],
-    [ 'N=CN=CCC', 'c:c:c:c:c:c:', 'quinazoline' ],
+    [ qw( NNCCCC CCCCCC cinnoline ) ],
+    [ qw( NCNCCC CCCCCC quinazoline ) ],
 
-    [ 'N=CC=NCC',     'c:c:c:c:c:c:', 'quinoxaline' ],
-    [ 'n:c:c:n:c:c:', 'c:c:c:c:c:c:', 'quinoxaline' ],
+    [ qw( NCCNCC CCCCCC quinoxaline ) ],
 
     [ 'c:c:c:n:c:c:', 'n:c:c:c:c:c:', '1,5-naphthyridine' ], # TODO: There are isomers
     [ 'c:n:n:c:c:c:', 'c:c:c:c:c:c:', 'phthalazine' ],
@@ -42,21 +43,23 @@ our @names = (
     [ 'C=NC=CCC',     'c:c:c:c:c:c:', 'isoquinoline' ],
     [ 'c:n:c:c:c:c:', 'c:c:c:c:c:c:', 'isoquinoline' ],
 
-    [ 'CC=CCNC',  'C=CC=CCN',  'quinolizine' ],
+    [ 'CC=CCn:c=', 'c:c:c:c:c:n:', 'quinolizine' ],
+    [ 'CC=CC=c:n', 'c:c:c:c:n:c:', 'quinolizine' ],
 
-    [ 'n:n:c:c:c:', 'c:c:c:c:c:c:', '1H-indazole' ],
-    [ 'n:c:c:c:c:', 'c:c:c:c:c:c:', '1H-indole' ],
+    [ 'n:n:c:c:c:', 'c:c:c:c:c:c:', 'indazole' ],
+    [ 'n:c:c:c:c:', 'c:c:c:c:c:c:', 'indole' ],
     [ 'c:n:c:c:c:', 'c:c:c:c:c:c:', 'isoindole' ],
-    [ 'c:c:c:c:n:', 'c:c:c:c:n:c:', 'indolizine', ],
-    [ 'CC=Cn:c',    'c:c:c:c:n:',   '1H-pyrrolizine' ], # TODO: There are isomers
+    [ 'c:c:c:c:c:n:', 'c:c:c:n:c:', 'indolizine', ],
+
+    [ qw( CCCCN CCCNC pyrrolizine ) ],
+    [ qw( CCCNC CCCNC pyrrolizine ) ],
 );
 
-for my $name (qw( 1H-indole indolizine isoindole isoquinoline quinoline quinolizine )) {
+for my $name (qw( indole indolizine isoindole isoquinoline quinoline quinolizine )) {
     for (grep { $_->[2] eq $name } @names) {
         my @As_parts = @$_;
         $As_parts[0] =~ s/n/\[as\]/g;
         $As_parts[1] =~ s/n/\[as\]/g;
-        $As_parts[2] =~ s/^1H-//;
         $As_parts[2] =~ s/^qu//;
         $As_parts[2] = 'ars' . $As_parts[2] unless $As_parts[2] =~ s/^iso(qu)?/isoars/;
         push @names, \@As_parts;
@@ -64,13 +67,18 @@ for my $name (qw( 1H-indole indolizine isoindole isoquinoline quinoline quinoliz
         my @P_parts = @$_;
         $P_parts[0] =~ s/n/p/g;
         $P_parts[1] =~ s/n/p/g;
-        $P_parts[2] =~ s/^1H-//;
         $P_parts[2] =~ s/^qu//;
         $P_parts[2] = 'phosph' . $P_parts[2] unless $P_parts[2] =~ s/^iso(qu)?/isophosph/;
         push @names, \@P_parts;
     }
 }
 
+for (@names) {
+    $_->[0] =~ s/[=:]//g;
+    $_->[1] =~ s/[=:]//g;
+    ( $_->[0], $_->[1] ) = map { uc } @$_[0..1];
+}
+
 # From BBv2 P-25.1.1, order of decreasing seniority
 our %hydrocarbons_by_size = (
     '5,7' => 'azulene',
@@ -88,15 +96,15 @@ sub new
     my $subgraph = $graph->subgraph( \@vertices );
     my @bridge = grep { $subgraph->degree( $_ ) == 3 } @vertices;
     $subgraph->delete_edge( @bridge );
-    $self->{vertices} = [ Graph::Traversal::DFS->new( $subgraph, start => $bridge[0] )->dfs ];
+    @vertices = Graph::Traversal::DFS->new( $subgraph, start => $bridge[0] )->dfs;
     $subgraph->delete_vertices( @bridge );
 
     # Graph is broken into components.
     # Each component is represented as an array of vertices in the order of traverse.
     my @components = sort { @$a <=> @$b } $subgraph->connected_components;
     for (0..1) {
-        if( @{$components[$_]} == 1 ) {
-            die "bicycles with three-membered cycles are not supported yet\n"
+        if( @{$components[$_]} < 3 ) {
+            die "bicycles with three and four-membered cycles are not supported yet\n"
         }
 
         my $subgraph = $graph->subgraph( [ @{$components[$_]}, @bridge ] );
@@ -111,13 +119,13 @@ sub new
 
     my $nbenzene = scalar grep { $_->is_benzene } @cycles;
 
-    # The ordering should not be done if one of the cycles is benzene
-    if( $nbenzene == 0 ) {
-        # The following code is supposed to order the rings _and_ establish the traversal order
-        # TODO: Maybe all traversals of both rings should be generated here?
-        # FIXME: There seem to be two orders for rings: one for numbering, other for fusion naming...
-        my @candidates = map { ChemOnomatopist::Chain::Monocycle->new( $_->graph, $_->vertices ) }
-                             ( @cycles, map { $_->flipped } @cycles );
+    if( !$nbenzene && $self->is_purine ) {
+        return ChemOnomatopist::Chain::Bicycle::Purine->new( $graph, @cycles );
+    } elsif( !$nbenzene ) {
+        # Find the senior cycle
+        my @candidates = map { $_->candidates }
+                         map { ChemOnomatopist::Chain::Monocycle->new( $_->graph, $_->vertices ) }
+                             $self->cycles;
         for my $rule ( # P-25.3.2.4 (a): Senior heteroatom according to specific seniority order
                        \&rule_most_senior_heteroatom,
                        # TODO: P-25.3.2.4 (b): Concerns fusions of more than two rings
@@ -140,32 +148,53 @@ sub new
             if( @candidates_now == 1 ) {
                 @candidates = @candidates_now;
                 last;
-            } elsif( @candidates ) {
+            } elsif( @candidates ) { # CHECKME: This looks strange
                 @candidates = @candidates_now;
             } else {
                 last;
             }
         }
+
         # Here the "winning" ring is selected
         my $chain = shift @candidates;
 
         # Making the "winning" ring the first
-        if( set( $chain->vertices ) == set( $cycles[1]->vertices ) ) {
-            @cycles = reverse @cycles;
+        if( set( $chain->vertices ) == set( $self->{cycles}[1]->vertices ) ) {
+            $self->{cycles} = [ reverse $self->cycles ];
         }
 
-        # Now we have to match the direction, I guess
-        my  @bridge_in_chain = grep { set( @bridge )->has( $_ ) } $chain->vertices;
-        if( $bridge_in_chain[-1] != $cycles[0]->{vertices}[-1] ) {
-            @cycles = map { $_->flipped } @cycles;
+        # Construct the candidates to determine the numbering order
+        @candidates = ( $self,
+                        $self->flipped_horizontally,
+                        $self->flipped_vertically,
+                        $self->flipped_horizontally->flipped_vertically );
+
+        # Establish the order
+        for my $rule (
+                       # P-25.3.3.1.2 (a): Lower locants for heteroatoms
+                       \&ChemOnomatopist::rule_lowest_numbered_heteroatoms,
+                       # P-25.3.3.1.2 (b): Lower locants for senior heteroatoms
+                       \&ChemOnomatopist::rule_lowest_numbered_most_senior_heteroatoms,
+                       # P-25.3.3.1.2 (c): Lower locants for fusion carbon atoms
+                       \&rule_lowest_numbered_fusion_carbons,
+                       # P-25.3.3.1.2 (d): Lower locants for fusion heteroatoms (rather than nonfusion)
+                       \&rule_lowest_numbered_fusion_heteroatoms,
+                       # TODO: P-25.3.3.1.2 (e): Lower locants for interior heteroatom
+                       # P-25.3.3.1.2 (f): Lower locants for indicated hydrogen atoms
+                       \&ChemOnomatopist::rule_lowest_numbered_indicated_hydrogens,
+                     ) {
+            my @candidates_now = $rule->( @candidates );
+            if( @candidates_now == 1 ) {
+                @candidates = @candidates_now;
+                last;
+            } elsif( @candidates ) { # CHECKME: This looks strange
+                @candidates = @candidates_now;
+            } else {
+                last;
+            }
         }
 
-        $self->{cycles} = \@cycles;
-        $self->_adjust_vertices_to_cycles;
-
-        if( $self->is_purine ) {
-            return ChemOnomatopist::Chain::Bicycle::Purine->new( $graph, @cycles );
-        }
+        $self->{vertices} = [ (shift @candidates)->vertices ]; # FIXME: Simply return instead of self
     } elsif( $nbenzene == 1 ) {
         # Numbering has to start from cycle other than benzene
         if( $cycles[0]->is_benzene ) {
@@ -180,7 +209,7 @@ sub new
             @cycles = map { $_->flipped } @cycles;
             $self->{cycles} = \@cycles;
         }
-        $self->_adjust_vertices_to_cycles;
+        $self->_adjust_vertices_to_cycles; # CHECKME: Is this needed?
     }
 
     return $self;
@@ -193,23 +222,41 @@ sub candidates()
 
     if( $self->is_naphthalene ) {
         # Generates all variants
-        push @chains, $self->copy, $self->copy, $self->copy;
-
-        $chains[1]->{cycles} = [ map { $_->flipped } $chains[1]->cycles ];
-        $chains[3]->{cycles} = [ map { $_->flipped } $chains[3]->cycles ];
-
-        $chains[2]->{cycles} = [ reverse $chains[2]->cycles ];
-        $chains[3]->{cycles} = [ reverse $chains[3]->cycles ];
-
-        for (@chains) {
-            $_->_adjust_vertices_to_cycles;
-            $_->{candidate_for} = $self unless $_ == $self;
+        push @chains, $self->flipped_horizontally,
+                      $self->flipped_vertically,
+                      $self->flipped_horizontally->flipped_vertically;
+        for (1..3) {
+            $chains[$_]->{candidate_for} = $self;
         }
     }
 
     return @chains;
 }
 
+sub flipped_horizontally()
+{
+    my( $self ) = @_;
+    my $copy = $self->copy;
+    my @vertices = reverse $copy->vertices;
+    push @vertices, shift @vertices;
+    $copy->{vertices} = \@vertices;
+    return $copy;
+}
+
+sub flipped_vertically()
+{
+    my( $self ) = @_;
+    my $copy = $self->copy;
+    my $cycle = first { set( $_->vertices )->has( $copy->{vertices}[0] ) }
+                      $copy->cycles;
+    my @vertices = reverse $copy->vertices;
+    for (1..$cycle->length-2) {
+        unshift @vertices, pop @vertices;
+    }
+    $copy->{vertices} = \@vertices;
+    return $copy;
+}
+
 sub copy()
 {
     my( $self ) = @_;
@@ -226,6 +273,28 @@ sub cycles()
     return @{$self->{cycles}};
 }
 
+sub fusion_positions()
+{
+    my( $self ) = @_;
+    my $bridge = set( map { @{$_->{vertices}}[-2..-1] } $self->cycles );
+    my @vertices = $self->vertices;
+    return grep { $bridge->has( $vertices[$_] ) } 0..$#vertices;
+}
+
+sub fusion_carbon_positions()
+{
+    my( $self ) = @_;
+    return grep { ChemOnomatopist::element( $self->{vertices}[$_] ) eq 'C' }
+                $self->fusion_positions;
+}
+
+sub fusion_heteroatom_positions()
+{
+    my( $self ) = @_;
+    return grep { ChemOnomatopist::element( $self->{vertices}[$_] ) ne 'C' }
+                $self->fusion_positions;
+}
+
 sub parent(;$)
 {
     my( $self, $parent ) = @_;
@@ -247,13 +316,17 @@ sub has_form($$)
     my %degrees = map { $graph->degree( $_ ) => 1 } $graph->vertices;
     return '' unless join( ',', sort keys %degrees ) eq '2,3';
 
+    # Safeguard against graphs such as bicyclo[3.2.1]octane
     my @d3 = grep { $graph->degree( $_ ) == 3 } $graph->vertices;
     return '' unless @d3 == 2;
     return '' unless $graph->has_edge( @d3 );
 
-    $graph = $graph->copy->delete_vertices( @d3 );
-    return '' unless scalar( $graph->connected_components ) == 2;
-    return 1;
+    # Safeguard against graphs such as 1,1'-biphenyl
+    $graph = $graph->copy->delete_edge( @d3 );
+    return '' unless $graph->is_connected;
+
+    $graph->delete_vertices( @d3 );
+    return !$graph->is_connected;
 }
 
 # Tells whether the outer bonds of the bicycle qualify as aromatic
@@ -320,9 +393,13 @@ sub needs_substituent_locants() { 1 }
 sub indicated_hydrogens_part()
 {
     my( $self ) = @_;
+    my $part = ChemOnomatopist::Name->new;
+    return $part unless $self->number_of_indicated_hydrogens;
+
     return $self->SUPER::indicated_hydrogens_part if $self->is_naphthalene;
 
-    my $part = ChemOnomatopist::Name->new;
+    # Use parent procedure if all the indicated hydrogens are confined to a single ring
+    return $self->SUPER::indicated_hydrogens_part if any { !$_->number_of_indicated_hydrogens } $self->cycles;
 
     if( $self->number_of_indicated_hydrogens &&
         $self->number_of_indicated_hydrogens < $self->length ) {
@@ -356,7 +433,7 @@ sub suffix()
 
     if( $self->is_hydrocarbon ) {
         # FIXME: Check if aromatic, but with caution, as substitutions will break aromaticity
-        my $cycle_sizes = join ',', map { $_->length } $self->cycles;
+        my $cycle_sizes = join ',', sort map { $_->length } $self->cycles;
         if( exists $hydrocarbons_by_size{$cycle_sizes} ) {
             return ChemOnomatopist::Name::Part::Stem->new( $hydrocarbons_by_size{$cycle_sizes} )->to_name;
         }
@@ -369,33 +446,30 @@ sub suffix()
 
     my @SMILES = map { $_->backbone_SMILES } $self->cycles;
     print STDERR "bicycle SMILES: @SMILES\n" if $ChemOnomatopist::DEBUG;
-    my $retained = first { ($_->[0] eq $SMILES[0] && $_->[1] eq $SMILES[1]) ||
-                           ($_->[0] eq $SMILES[1] && $_->[1] eq $SMILES[0]) } @names;
+    my $retained = find_retained( @SMILES );
     return ChemOnomatopist::Name::Part::Stem->new( $retained->[2] )->to_name if $retained;
 
     if( any { $_->is_benzene } $self->cycles ) {
         my $other = first { !$_->is_benzene } $self->cycles;
 
-        my $SMILES = ChemOnomatopist::Chain::Monocycle->new( $other->graph, $other->vertices )->backbone_SMILES;
-        if( $SMILES =~ /^C=C((?<el>O|S|\[Se\]|\[Te\])C|C(?<el>O|S|\[Se\]|\[Te\]))c:c$/ ) {
-            # Names according to BBv2 P-25.2.1, Table 2.8, (23) and (24)
-            my $element = $+{el};
-            my $name = 'benzo';
-            $element =~ s/[\[\]]//g;
+        my $name = ChemOnomatopist::Name->new( 'benzo' );
+        if( $other->length == 6 && $other->is_monoreplaced &&
+            join( '', $other->heteroatoms ) =~ /^(O|S|Se|Te)$/ &&
+            join( '', $other->heteroatom_positions ) < 4 ) {
+            # Names according to BBv3 P-25.2.1, Table 2.8, (23) and (24)
+            my( $element ) = $other->heteroatoms;
             if( $element ne 'O' ) {
                 $name .= $elements{$element}->{prefix};
-                $name =~ s/a$/o/;
+                $name->[-1] =~ s/a$/o/;
             }
             return $name . 'pyran';
         } else {
-            my $name = ChemOnomatopist::Name->new( 'benzo' );
             my  $other_name = $other->suffix;
             if( $other_name->starts_with_locant ) { # Locants are moved to front
                 unshift @$name, shift @$other_name;
             }
             $name->[-1] =~ s/o$// if $other_name->[0] =~ /^a/;
-            $name .= $other_name;
-            return $name;
+            return $name . $other_name;
         }
     }
 
@@ -415,46 +489,58 @@ sub suffix()
     my $min_B = min map {  $_->vertex_ids( @bridge ) } @equiv_B;
     @equiv_B = grep { min( $_->vertex_ids( @bridge ) ) == $min_B } @equiv_B;
 
-    my $fusion = '[';
-    if( @equiv_A > 1 || @equiv_B > 1 ) {
+    my $fusion;
+    if( $self->{cycles}[1]->is_homogeneous &&
+        $self->{cycles}[1]->number_of_branches == 2 ) {
+        $fusion = '[';
+    } elsif( @equiv_A > 1 || @equiv_B > 1 ) {
         # At least one of the rings has mirror symmetry ("flip-symmetric"), thus numeric order is ascending
-        $fusion .= ($min_B+1) . ',' . ($min_B+2);
+        $fusion = '[' . ($min_B+1) . ',' . ($min_B+2) . '-';
     } else {
         # Rings are rigid, thus numeric order has to be derived
         my @order_A = $equiv_A[0]->vertex_ids( @bridge );
         my @order_B = $equiv_B[0]->vertex_ids( @bridge );
         if( ($order_A[0] <=> $order_A[1]) == ($order_B[0] <=> $order_B[1]) ) {
             # Ring atoms are encountered in the same order in both of the rings
-            $fusion .= ($min_B+1) . ',' . ($min_B+2);
+            $fusion = '[' . ($min_B+1) . ',' . ($min_B+2);
         } else {
             # Ring atom orders differ
-            $fusion .= ($min_B+2) . ',' . ($min_B+1);
+            $fusion = '[' . ($min_B+2) . ',' . ($min_B+1);
         }
+        $fusion .= '-';
     }
-    $fusion .= '-' . chr( 97 + $min_A ) . ']';
-
-    my $name = ChemOnomatopist::Name->new;
-    my $graph = $self->graph;
+    $fusion .= chr( 97 + $min_A ) . ']';
 
     my @ideal = map { ChemOnomatopist::Chain::Monocycle->new( $_->graph, $_->vertices ) }
                     $self->cycles;
-    my $name_A = $ideal[1]->name;
+    my $name_A = $ideal[1]->suffix;
     $name_A =~ s/^\d+H-//;
 
-    # TODO: Complete retained prefixes from BBv2 P-25.3.2.2.3
-    $name_A = 'fur'     if $name_A eq 'furan';
-    $name_A = 'imidaz'  if $name_A eq 'imidazole';
-    $name_A = 'pyrid'   if $name_A eq 'pyridine';
-    $name_A = 'pyrimid' if $name_A eq 'pyrimidine';
-    $name_A = 'thien'   if $name_A eq 'thiophene';
+    # Retained prefixes from BBv3 P-25.3.2.2.3
+    $name_A = 'anthra'    if $name_A eq 'anthracene';
+    $name_A = 'naphth'    if $name_A eq 'naphthalene';
+    $name_A = 'benz'      if $name_A eq 'benzene';
+    $name_A = 'phenanthr' if $name_A eq 'phenanthrene';
+    $name_A = 'fur'       if $name_A eq 'furan';
+    $name_A = 'imidaz'    if $name_A eq 'imidazole';
+    $name_A = 'pyrid'     if $name_A eq 'pyridine';
+    $name_A = 'pyrimid'   if $name_A eq 'pyrimidine';
+    $name_A = 'thien'     if $name_A eq 'thiophene';
 
-    $name .= $name_A;
-    unless( $name->[-1] =~ s/e$/o/ ) { # BBv2 P-25.3.2.2.2
-        $name->[-1] .= 'o';
+    $name_A = ChemOnomatopist::Name->new( $name_A ) unless blessed $name_A;
+    $name_A->pop_e;
+    if( $name_A->ends_with_alkane_an_suffix ) {
+        pop @$name_A;
+        $name_A .= 'a';
+    } else {
+        $name_A .= 'o';
     }
+
+    my $name = ChemOnomatopist::Name->new;
+    $name .= $name_A;
     $name .= ChemOnomatopist::Name::Part::Fusion->new( $fusion );
 
-    my $name_B = $ideal[0]->name;
+    my $name_B = $ideal[0]->suffix;
     $name_B->[0] =~ s/\d+H-//;
     $name .= $name_B;
     $name->bracket_numeric_locants;
@@ -476,14 +562,11 @@ sub rule_most_senior_heteroatom
     return grep { any { $_ eq $max_value } $_->heteroatoms } @chains;
 }
 
-sub rule_greatest_variety_of_heteroatoms
-{
-    my( @chains ) = @_;
+sub rule_greatest_variety_of_heteroatoms { all_max { scalar uniq $_->heteroatoms } @_ }
 
-    my( $max_value ) = reverse sort map { scalar uniq $_->heteroatoms } @chains;
-    return @chains unless $max_value;
-    return grep { scalar( uniq $_->heteroatoms ) == $max_value } @chains;
-}
+sub rule_lowest_numbered_fusion_carbons { all_min { ChemOnomatopist::Comparable::Array::Numeric->new( $_->fusion_carbon_positions ) } @_ }
+
+sub rule_lowest_numbered_fusion_heteroatoms { all_min { ChemOnomatopist::Comparable::Array::Numeric->new( $_->fusion_heteroatom_positions ) } @_ }
 
 sub _adjust_vertices_to_cycles()
 {
@@ -500,4 +583,11 @@ sub _adjust_vertices_to_cycles()
     return $self;
 }
 
+sub find_retained
+{
+    my @SMILES = map { s/[=:]//g; uc } @_;
+    return first { ($_->[0] eq $SMILES[0] && $_->[1] eq $SMILES[1]) ||
+                   ($_->[0] eq $SMILES[1] && $_->[1] eq $SMILES[0]) } @names;
+}
+
 1;
