@@ -9,7 +9,7 @@ use warnings;
 use ChemOnomatopist::Elements qw( %elements );
 use ChemOnomatopist::Group::Amide;
 use ChemOnomatopist::Name;
-use List::Util qw( first );
+use List::Util qw( any first );
 use Scalar::Util qw( blessed );
 
 use parent ChemOnomatopist::Group::, ChemOnomatopist::Chain::;
@@ -18,6 +18,9 @@ sub new
 {
     my( $class, $graph, @vertices ) = @_;
     my $ketone = first { blessed $_ && $_->isa( ChemOnomatopist::Group::Ketone:: ) } @vertices;
+    if( $graph->degree( $vertices[3] ) > $graph->degree( $vertices[2] ) ) {
+        @vertices[2..3] = reverse @vertices[2..3];
+    }
     return bless { graph => $graph,
                    vertices => \@vertices,
                    ketone_element => $ketone->{element} },
@@ -26,7 +29,13 @@ sub new
 
 sub heteroatom_positions() { my @not_applicable }
 
-sub needs_substituent_locants() { $_[0]->number_of_branches > 1 && $_[0]->number_of_branches < 4 }
+sub needs_substituent_locants()
+{
+    my( $self ) = @_;
+    return 1 if $self->number_of_branches > 1 && $self->number_of_branches < 4;
+    return 1 if any { $_->has_substituent_locant } map { @$_ } $self->locant_names;
+    return '';
+}
 
 sub locants(@)
 {
